/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the OX Software GmbH group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016-2020 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.filestore;

import com.openexchange.exception.Category;
import com.openexchange.exception.DisplayableOXExceptionCode;
import com.openexchange.exception.OXException;
import com.openexchange.exception.OXExceptionFactory;
import com.openexchange.exception.OXExceptionStrings;

/**
 * Error codes for the file storage exception.
 *
 * @author Steffen Templin
 */
public enum QuotaFileStorageExceptionCodes implements DisplayableOXExceptionCode {
    /**
     * Couldn't reach the filestore
     */
    INSTANTIATIONERROR("File store could not be accessed.", OXExceptionStrings.MESSAGE, Category.CATEGORY_SERVICE_DOWN, 21),
    /**
     * Database Query could not be realized
     */
    SQLSTATEMENTERROR("Database query failed.", OXExceptionStrings.SQL_ERROR_MSG, Category.CATEGORY_ERROR, 23),
    /**
     * The allowed Quota is reached.
     */
    STORE_FULL("The allowed Quota is reached.", QuotaFileStorageExceptionMessage.STORE_FULL_MSG, Category.CATEGORY_USER_INPUT, 24),
    /**
     * Quota seems to be inconsistent. Please use consistency tool for context %1$s.
     */
    QUOTA_UNDERRUN("Quota seems to be inconsistent. Please use consistency tool for context %1$s.", OXExceptionStrings.MESSAGE, Category.CATEGORY_TRUNCATED, 25),
    /**
     * Quota seems to be inconsistent. Please use consistency tool for owner %1$s in context %2$s.
     */
    QUOTA_UNDERRUN_USER("Quota seems to be inconsistent. Please use consistency tool for owner %1$s in context %2$s.", OXExceptionStrings.MESSAGE, Category.CATEGORY_TRUNCATED, 25),
    /**
     * Quota usage is missing for context %1$s.
     */
    NO_USAGE("Quota usage is missing for context %1$s.", OXExceptionStrings.MESSAGE, Category.CATEGORY_ERROR, 26),
    /**
     * Quota usage is missing for owner %1$s in context %2$s.
     */
    NO_USAGE_USER("Quota usage is missing for owner %1$s in context %2$s.", OXExceptionStrings.MESSAGE, Category.CATEGORY_ERROR, 26),
    /**
     * Update of quota usage for context %1$s failed.
     */
    UPDATE_FAILED("Update of quota usage for context %1$s failed.", OXExceptionStrings.MESSAGE, Category.CATEGORY_ERROR, 27),
    /**
     * Update of quota usage for owner %1$s context %2$s failed.
     */
    UPDATE_FAILED_USER("Update of quota usage for owner %1$s in context %2$s failed.", OXExceptionStrings.MESSAGE, Category.CATEGORY_ERROR, 27),
    ;

    private final String message;
    private final String displayMessage;
    private final Category category;
    private final int number;

    /**
     * Initializes a new {@link QuotaFileStorageExceptionCodes}.
     */
    private QuotaFileStorageExceptionCodes(final String message, String displayMessage, final Category category, final int detailNumber) {
        this.message = message;
        this.displayMessage = displayMessage;
        this.category = category;
        this.number = detailNumber;
    }

    @Override
    public String getPrefix() {
        return FileStorageCodes.prefix();
    }

    @Override
    public Category getCategory() {
        return category;
    }

    @Override
    public int getNumber() {
        return number;
    }

    @Override
    public String getMessage() {
        return message;
    }

    @Override
    public String getDisplayMessage() {
        return displayMessage;
    }

    @Override
    public boolean equals(final OXException e) {
        return OXExceptionFactory.getInstance().equals(this, e);
    }

    /**
     * Creates a new {@link OXException} instance pre-filled with this code's attributes.
     *
     * @return The newly created {@link OXException} instance
     */
    public OXException create() {
        return OXExceptionFactory.getInstance().create(this, new Object[0]);
    }

    /**
     * Creates a new {@link OXException} instance pre-filled with this code's attributes.
     *
     * @param args The message arguments in case of printf-style message
     * @return The newly created {@link OXException} instance
     */
    public OXException create(final Object... args) {
        return OXExceptionFactory.getInstance().create(this, (Throwable) null, args);
    }

    /**
     * Creates a new {@link OXException} instance pre-filled with this code's attributes.
     *
     * @param cause The optional initial cause
     * @param args The message arguments in case of printf-style message
     * @return The newly created {@link OXException} instance
     */
    public OXException create(final Throwable cause, final Object... args) {
        return OXExceptionFactory.getInstance().create(this, cause, args);
    }

}
