/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the OX Software GmbH. group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016-2020 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.groupware.update;

import java.sql.Connection;
import java.sql.SQLException;
import com.openexchange.database.Databases;
import com.openexchange.exception.OXException;


/**
 * {@link TransactionalUpdateTaskAdapter}
 *
 * @author <a href="mailto:thorben.betten@open-xchange.com">Thorben Betten</a>
 * @since v7.10.0
 */
public abstract class TransactionalUpdateTaskAdapter extends UpdateTaskAdapter {

    /**
     * Initializes a new {@link TransactionalUpdateTaskAdapter}.
     */
    public TransactionalUpdateTaskAdapter() {
        super();
    }

    @Override
    public final void perform(PerformParameters params) throws OXException {
        Connection con = params.getConnection();
        try {
            if (false == needsChanges(con)) {
                // Nothing needs to be done
                return;
            }

            // Perform changes in transaction
            boolean rollback = false;
            try {
                con.setAutoCommit(false);
                rollback = true;

                performChanges(params, con);

                con.commit();
                rollback = false;
            } finally {
                if (rollback) {
                    Databases.rollback(con);
                }
                Databases.autocommit(con);
            }
        } catch (SQLException e) {
            throw UpdateExceptionCodes.SQL_PROBLEM.create(e, e.getMessage());
        } catch (RuntimeException e) {
            throw UpdateExceptionCodes.OTHER_PROBLEM.create(e, e.getMessage());
        }
    }

    /**
     * Performs the changes for this update task. Given connection is in transaction.
     *
     * @param params The parameters carrying some useful parameters for performing the update.
     * @param con The connection to use
     * @throws OXException If an Open-Xchange error occurs
     * @throws SQLException If an SQL error occurs
     */
    protected abstract void performChanges(PerformParameters params, Connection con) throws OXException, SQLException;

    /**
     * Checks whether changes are required to be performed by this update task
     *
     * @param con The connection to use
     * @return <code>true</code> if changes are required; otherwise <code>false</code> for no-op
     * @throws SQLException If an SQL error occurs
     */
    protected boolean needsChanges(Connection con) throws SQLException{
        return true;
    }

}
