/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the OX Software GmbH group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016-2020 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.sessionstorage.hazelcast.serialization;

import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import org.apache.commons.codec.DecoderException;
import org.apache.commons.codec.net.URLCodec;
import com.hazelcast.nio.serialization.ClassDefinition;
import com.hazelcast.nio.serialization.ClassDefinitionBuilder;
import com.hazelcast.nio.serialization.PortableReader;
import com.hazelcast.nio.serialization.PortableWriter;
import com.hazelcast.nio.serialization.VersionedPortable;
import com.openexchange.hazelcast.serialization.CustomPortable;
import com.openexchange.java.Charsets;
import com.openexchange.java.StringAppender;
import com.openexchange.session.Session;
import com.openexchange.sessionstorage.SessionStorageConfiguration;
import com.openexchange.sessionstorage.StoredSession;

/**
 * {@link PortableSession} - The portable representation for {@link Session} type.
 *
 * @author <a href="mailto:tobias.friedrich@open-xchange.com">Tobias Friedrich</a>
 */
public class PortableSession extends StoredSession implements CustomPortable, VersionedPortable {

    private static final long serialVersionUID = -2346327568417617677L;

    private static final org.slf4j.Logger LOG = org.slf4j.LoggerFactory.getLogger(PortableSession.class);

    /**
     * BitSet of www-form-url safe characters.
     */
    protected static final BitSet BIT_SET_PARAMS;

    // Static initializer for BIT_SET_PARAMS
    static {
        final BitSet bitSet = new BitSet(256);

        // Exclude ':' and '%' from printable ASCII characters
        for (int i = 32; i < 37; i++) {
            bitSet.set(i);
        }
        for (int i = 38; i < 58; i++) {
            bitSet.set(i);
        }
        for (int i = 59; i < 127; i++) {
            bitSet.set(i);
        }

        BIT_SET_PARAMS = bitSet;
    }

    /** The unique portable class ID of the {@link PortableSession} */
    public static final int CLASS_ID = 1;

    /**
     * The class version for {@link PortableSession}
     * <p>
     * This number should be incremented whenever fields are added;
     * see <a href="http://docs.hazelcast.org/docs/latest-development/manual/html/Serialization/Implementing_Portable_Serialization/Versioning_for_Portable_Serialization.html">here</a> for reference.
     */
    public static final int CLASS_VERSION = 2;

    public static final String PARAMETER_LOGIN_NAME = "loginName";
    public static final String PARAMETER_PASSWORD = "password";
    public static final String PARAMETER_CONTEXT_ID = "contextId";
    public static final String PARAMETER_USER_ID = "userId";
    public static final String PARAMETER_SESSION_ID = "sessionId";
    public static final String PARAMETER_SECRET = "secret";
    public static final String PARAMETER_LOGIN = "login";
    public static final String PARAMETER_RANDOM_TOKEN = "randomToken";
    public static final String PARAMETER_LOCAL_IP = "localIp";
    public static final String PARAMETER_AUTH_ID = "authId";
    public static final String PARAMETER_HASH = "hash";
    public static final String PARAMETER_CLIENT = "client";
    public static final String PARAMETER_USER_LOGIN = "userLogin";
    public static final String PARAMETER_ALT_ID = "altId";
    public static final String PARAMETER_USER_AGENT = "userAgent";
    public static final String PARAMETER_LOGIN_TIME = "loginTime";
    public static final String PARAMETER_LOCAL_LAST_ACTIVE = "localLastActive";
    public static final String PARAMETER_REMOTE_PARAMETER_NAMES = "remoteParameterNames";
    public static final String PARAMETER_REMOTE_PARAMETER_VALUES = "remoteParameterValues";

    /** The class definition for PortableSession */
    public static ClassDefinition CLASS_DEFINITION = new ClassDefinitionBuilder(FACTORY_ID, CLASS_ID, CLASS_VERSION)
        .addUTFField(PARAMETER_LOGIN_NAME)
        .addUTFField(PARAMETER_PASSWORD)
        .addIntField(PARAMETER_CONTEXT_ID)
        .addIntField(PARAMETER_USER_ID)
        .addUTFField(PARAMETER_SESSION_ID)
        .addUTFField(PARAMETER_SECRET)
        .addUTFField(PARAMETER_LOGIN)
        .addUTFField(PARAMETER_RANDOM_TOKEN)
        .addUTFField(PARAMETER_LOCAL_IP)
        .addUTFField(PARAMETER_AUTH_ID)
        .addUTFField(PARAMETER_HASH)
        .addUTFField(PARAMETER_CLIENT)
        .addUTFField(PARAMETER_USER_LOGIN)
        .addUTFField(PARAMETER_ALT_ID)
        .addUTFField(PARAMETER_USER_AGENT)
        .addLongField(PARAMETER_LOGIN_TIME)
        .addLongField(PARAMETER_LOCAL_LAST_ACTIVE)
        .addUTFField(PARAMETER_REMOTE_PARAMETER_NAMES)
        .addUTFField(PARAMETER_REMOTE_PARAMETER_VALUES)
        .build();

    // -------------------------------------------------------------------------------------------------

    private Set<String> remoteParameterNames;
    private Long localLastActive;

    /**
     * Initializes a new {@link PortableSession}.
     */
    public PortableSession() {
        super();
        localLastActive = null;
        remoteParameterNames = Collections.emptySet();
    }

    /**
     * Initializes a new {@link PortableSession}.
     *
     * @param session The underlying session
     */
    public PortableSession(final Session session) {
        super(session);
        localLastActive = null;
        Collection<String> configuredRemoteParameterNames = SessionStorageConfiguration.getInstance().getRemoteParameterNames(userId, contextId);
        Set<String> remoteParameterNames = new LinkedHashSet<>(configuredRemoteParameterNames.size() + 2); // Keep order

        // Add static remote parameters
        remoteParameterNames.add(PARAM_OAUTH_ACCESS_TOKEN);

        // Add configured remote parameters
        remoteParameterNames.addAll(configuredRemoteParameterNames);
        this.remoteParameterNames = remoteParameterNames;
    }

    /**
     * Sets the local last-active time stamp.
     *
     * @param localLastActive The local last-active time stamp to set
     */
    public void setLocalLastActive(long localLastActive) {
        this.localLastActive = Long.valueOf(localLastActive);
    }

    /**
     * Gets the local last-active time stamp.
     * <p>
     * <div style="margin-left: 0.1in; margin-right: 0.5in; margin-bottom: 0.1in; background-color:#FFDDDD;">
     * <b>Note</b>: This method might return <code>null</code>. Beware of auto-unboxing!
     * </div>
     *
     * @return The local last-active time stamp or <code>null</code> if not available
     */
    public Long getLocalLastActive() {
        return localLastActive;
    }

    @Override
    public int getFactoryId() {
        return FACTORY_ID;
    }

    @Override
    public int getClassId() {
        return CLASS_ID;
    }

    @Override
    public int getClassVersion() {
        return CLASS_VERSION;
    }

    @Override
    public void writePortable(PortableWriter writer) throws IOException {
        /*
         * basic properties
         */
        writer.writeUTF(PARAMETER_LOGIN_NAME, loginName);
        writer.writeUTF(PARAMETER_PASSWORD, password);
        writer.writeInt(PARAMETER_CONTEXT_ID, contextId);
        writer.writeInt(PARAMETER_USER_ID, userId);
        writer.writeUTF(PARAMETER_SESSION_ID, sessionId);
        writer.writeUTF(PARAMETER_SECRET, secret);
        writer.writeUTF(PARAMETER_LOGIN, login);
        writer.writeUTF(PARAMETER_RANDOM_TOKEN, randomToken);
        writer.writeUTF(PARAMETER_LOCAL_IP, localIp);
        writer.writeUTF(PARAMETER_AUTH_ID, authId);
        writer.writeUTF(PARAMETER_HASH, hash);
        writer.writeUTF(PARAMETER_CLIENT, client);
        writer.writeUTF(PARAMETER_USER_LOGIN, userLogin);
        /*
         * special handling for parameters map
         */
        {
            Object altId = parameters.get(PARAM_ALTERNATIVE_ID);
            writer.writeUTF(PARAMETER_ALT_ID, null == altId ? null : altId.toString());
        }
        {
            Object userAgent = parameters.get(PARAM_USER_AGENT);
            writer.writeUTF(PARAMETER_USER_AGENT, null == userAgent ? null : userAgent.toString());
        }
        {
            Object loginTime = parameters.get(PARAM_LOGIN_TIME);
            writer.writeLong(PARAMETER_LOGIN_TIME, null != loginTime ? ((Long) loginTime).longValue() : -1L);
        }
        {
            Long localLastActive = this.localLastActive;
            writer.writeLong(PARAMETER_LOCAL_LAST_ACTIVE, null != localLastActive ? localLastActive.longValue() : -1L);
        }
        {
            Set<String> remoteParameterNames = this.remoteParameterNames;
            StringAppender names = null;
            StringAppender values = null;
            for (String parameterName : remoteParameterNames) {
                Object value = parameters.get(parameterName);
                if (null != value) {
                    if (isSerializablePojo(value)) {
                        String sValue = value.toString();
                        if (null == names) {
                            int capacity = remoteParameterNames.size() << 4;
                            names = new StringAppender(':', capacity);
                            values = new StringAppender(':', capacity);
                        }
                        names.append(parameterName);
                        values.append(getSafeValue(sValue));
                        LOG.debug("Put remote parameter '{}' with value '{}' into portable session {} ({}@{})", parameterName, value, sessionId, Integer.valueOf(userId), Integer.valueOf(contextId));
                    } else {
                        LOG.warn("Denied remote parameter for name {}. Seems to be no ordinary Java object.", value.getClass().getName());
                    }
                }
            }
            if (null == names || null == values) {
                writer.writeUTF(PARAMETER_REMOTE_PARAMETER_NAMES, null);
                writer.writeUTF(PARAMETER_REMOTE_PARAMETER_VALUES, null);
            } else {
                writer.writeUTF(PARAMETER_REMOTE_PARAMETER_NAMES, names.toString());
                writer.writeUTF(PARAMETER_REMOTE_PARAMETER_VALUES, values.toString());
            }
        }
    }

    @Override
    public void readPortable(PortableReader reader) throws IOException {
        /*
         * basic properties
         */
        loginName = reader.readUTF(PARAMETER_LOGIN_NAME);
        password = reader.readUTF(PARAMETER_PASSWORD);
        contextId = reader.readInt(PARAMETER_CONTEXT_ID);
        userId = reader.readInt(PARAMETER_USER_ID);
        sessionId = reader.readUTF(PARAMETER_SESSION_ID);
        secret = reader.readUTF(PARAMETER_SECRET);
        login = reader.readUTF(PARAMETER_LOGIN);
        randomToken = reader.readUTF(PARAMETER_RANDOM_TOKEN);
        localIp = reader.readUTF(PARAMETER_LOCAL_IP);
        authId = reader.readUTF(PARAMETER_AUTH_ID);
        hash = reader.readUTF(PARAMETER_HASH);
        client = reader.readUTF(PARAMETER_CLIENT);
        userLogin = reader.readUTF(PARAMETER_USER_LOGIN);
        /*
         * special handling for parameters map
         */
        {
            String altId = reader.readUTF(PARAMETER_ALT_ID);
            if (null != altId) {
                parameters.put(PARAM_ALTERNATIVE_ID, altId);
            }
        }
        {
            String userAgent = reader.readUTF(PARAMETER_USER_AGENT);
            if (null != userAgent) {
                parameters.put(PARAM_USER_AGENT, userAgent);
            }
        }
        {
            long loginTime = reader.readLong(PARAMETER_LOGIN_TIME);
            if (loginTime > 0) {
                parameters.put(PARAM_LOGIN_TIME, Long.valueOf(loginTime));
            }
        }
        {
            long localLastActive = reader.readLong(PARAMETER_LOCAL_LAST_ACTIVE);
            this.localLastActive = localLastActive > 0 ? Long.valueOf(localLastActive) : null;
        }
        {
            String sNames = reader.readUTF(PARAMETER_REMOTE_PARAMETER_NAMES);
            if (null != sNames) {
                List<String> names = parseColonString(sNames);
                List<String> values = parseColonString(reader.readUTF(PARAMETER_REMOTE_PARAMETER_VALUES)); // Expect them, too
                Set<String> remoteParameterNames = new LinkedHashSet<>(names.size()); // Keep order
                for (int i = 0, size = names.size(); i < size; i++) {
                    try {
                        Object value = parseToSerializablePojo(decodeSafeValue(values.get(i)));
                        String name = names.get(i);
                        parameters.put(name, value);
                        remoteParameterNames.add(name);
                    } catch (DecoderException e) {
                        // Ignore
                    }
                }
                this.remoteParameterNames = remoteParameterNames;
            }
        }
    }

    private static List<String> parseColonString(String str) {
        List<String> retval = new ArrayList<String>();
        int length = str.length();
        {
            int prev = 0;
            int pos;
            while (prev < length && (pos = str.indexOf(':', prev)) >= 0) {
                if (pos > 0) {
                    retval.add(str.substring(prev, pos));
                }
                prev = pos + 1;
            }
            if (prev < length) {
                retval.add(str.substring(prev));
            }
        }
        return retval;
    }

    private static String getSafeValue(String sValue) {
        return sValue.indexOf(':') < 0 && sValue.indexOf('%') < 0 ? sValue : Charsets.toAsciiString(URLCodec.encodeUrl(BIT_SET_PARAMS, sValue.getBytes(Charsets.UTF_8)));
    }

    private static String decodeSafeValue(String value) throws DecoderException {
        return value.indexOf('%') < 0 ? value : new String(URLCodec.decodeUrl(Charsets.toAsciiBytes(value)), Charsets.UTF_8);
    }

    private static final String POJO_PACKAGE = "java.lang.";

    private static boolean isSerializablePojo(Object obj) {
        return null == obj ? false : ((obj instanceof Serializable) && obj.getClass().getName().startsWith(POJO_PACKAGE));
    }

    private static Object parseToSerializablePojo(String value) {
        if ("true".equalsIgnoreCase(value)) {
            return Boolean.TRUE;
        }
        if ("false".equalsIgnoreCase(value)) {
            return Boolean.FALSE;
        }

        try {
            int i = Integer.parseInt(value, 10);
            return Integer.valueOf(i);
        } catch (Exception e) {
            // Ignore...
        }

        try {
            long l = Long.parseLong(value, 10);
            return Long.valueOf(l);
        } catch (Exception e) {
            // Ignore...
        }

        /*-
         *
        try {
            float f = Float.parseFloat(value);
            return new Float(f);
        } catch (Exception e) {
            // Ignore...
        }

        try {
            double d = Double.parseDouble(value);
            return new Double(d);
        } catch (Exception e) {
            // Ignore...
        }
         *
         */

        return value;
    }

}
